/*
 * cQueue.cpp
 * 
 * Class implementatino for generic Fifo queue table.
 *
 * Authors:  Vicky & Ted
 */
#include "cQueue.h"
#include <assert.h>

#ifndef NULL
	#define NULL 0
#endif

/*
 * cQueue::cQueue()
 *
 * Purpose:	Create a new queue (circular, fifo).
 * IN:		size	-> The starting size of the queue.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The queue is set up.
 * Throws:	cQueueException
 * Return:	-
 */
cQueue::cQueue(int size)
{
	mSize			= size;
	mNumElements	= 0;
	mInsertPtr		= 0;
	mRemovePtr		= 0;

	// Alloc queue space.
	mQueue = new void*[size];
	if(!mQueue)
	{
		throw cQueueException("<cQueue::cQueue>: Unable to alloc new queue.");
	}
}

/*
 * cQueue::~cQueue()
 *
 * Purpose:	Destroy the queue.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	The queue is cleaned up.
 * Return:	-
 */
cQueue::~cQueue()
{
	mCS.Enter();
	if(mQueue) { delete[] mQueue; }
	mCS.Leave();
}

/*
 * cQueue::Push()
 *
 * Purpose:	Inserts the data into the front of the queue.
 * IN:		data	-> A pointer to the data object to be inserted.
 * OUT:		-
 * Cond:	The queue is not full.
 * PostCnd:	The free space in queue decreases by one.
 * Return:	true if success, else false.
 */ 
bool cQueue::Push(void *data)
{
	bool	retVal;
	mCS.Enter();

	if(mNumElements < mSize)
	{
		mQueue[mInsertPtr] = data;
		mNumElements++;
		mInsertPtr = (mInsertPtr +1) % mSize;
		retVal = true;
	}
	else
	{
		retVal = false;
	}

	mCS.Leave();
	return retVal;
}

/*
 * cQueue::Pop()
 *
 * Purpose:	Removes the next data element in fifo order.
 * IN:		-
 * OUT:		data	-> The data that was contained in the front node.
 * Cond:	There is data in the queue.
 * PostCnd:	The number of elements in the queue decreases.
 * Return:	true if success, else fail.
 */ 
bool cQueue::Pop(void **data)
{
	bool	retVal;
	mCS.Enter();

	if(mNumElements > 0)
	{
		*data = mQueue[mRemovePtr];
		mNumElements--;
		mRemovePtr = (mRemovePtr + 1) % mSize;
	}
	else
	{
		retVal = false;
	}

	mCS.Leave();
	return retVal;
}

/*
 * cQueue::Peek()
 *
 * Purpose:	Peeks at first element in queue but doesn't remove.
 * IN:		-
 * OUT:		data	-> The data that was contained in the front node.
 * Cond:	There is data in the queue.
 * PostCnd:	-
 * Return:	true if success, else fail.
 */ 
bool cQueue::Peek(void **data)
{
	bool retVal;
	mCS.Enter();
	
	if(mNumElements > 0)
	{
		*data = mQueue[mRemovePtr];
		retVal = true;
	}
	else
	{
		retVal = false;
	}

	mCS.Leave();
	return retVal;
}

/*
 * cQueue::Clear()
 *
 * Purpose:	Clears out and resets the queue.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	The queue is reset and cleared.
 * Return:	true if success, false otherwise.
 */
bool cQueue::Clear()
{
	mCS.Enter();

	mNumElements = 0;
	mInsertPtr   = 0;
	mRemovePtr   = 0;

	mCS.Leave();
	return true;
}
