/* 
 * cS_Factory.cpp
 * Class implementation for shared factory.
 *
 * Used to create reader and writer shared containers.
 *
 */
#include "cS_Factory.h"
#include "cHashTable.h"
#include <iostream.h>

void AM_SetErrorStream(ostream &errStream);

// Set up static members
int					cS_Factory::mRefCount = 0;
eb_t				cS_Factory::mBundle   = NULL;
cCriticalSection	cS_Factory::mCS;
cHashTable*			cS_Factory::mWriterTable;
cHashTable*			cS_Factory::mReaderTable;
cMutex				cS_Factory::mRegisterMutex;
cMutex				cS_Factory::mUnregisterMutex;
volatile bool		cS_Factory::mRegistered;
volatile bool		cS_Factory::mUnregistered;

/*
 * cS_Factory::CreateReader()
 * 
 * Purpose:	Attempts to create a new reader, listening on the given container.
 * IN:		originalCont	-> Pointer to the original container.
 *			writer			-> The name of the writer to register with.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The mReader table has one more reader in it?
 * Return:	true (always, heh!)
 */
cS_Reader* cS_Factory::CreateReader(cStdCont *originalCont, char *writer, Callback callbackFunction)
{
	_en_t		writerName;
	en_t		newEndpointName;
	ep_t		newEndpoint;
	cS_Reader  *newReader;
	bool		result;

	if(!_IncreaseRefCount())
	{
		return NULL;
	}

	// Allocate an endpoint for this reader.
	if(AM_AllocateEndpoint(mBundle, &newEndpoint, &newEndpointName) != AM_OK)
	{
		cout << "Unable to allocate endpoint in CreateReader." << endl;
		return NULL;
	}
	AM_SetTag(newEndpoint, S_TAG);	// Set it to the secret tag
	if(AM_SetExpectedResources(newEndpoint, 10, 10) != AM_OK)
	{
		cout << "Error: Unable to allocate enough endpoint resources in CreateReader." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Lookup the writer's address.
	if(AMNS_Lookup(writer, &writerName) != AM_OK)
	{
		cout << "Unable to lookup endpoint name in CreateReader." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Map the writer.
	if(AM_Map(newEndpoint, S_WRITERMAP, &writerName, S_TAG) != AM_OK)
	{
		cout << "Unable to map the writer's endpoint for register." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Map the RegisterReply handler.
	if(AM_SetHandler(newEndpoint, S_REGISTER_REPLY_HANDLER, (Function)cS_Factory::RegisterReplyHandler) != AM_OK)
	{
		cout << "Error: Unable to set the register reply handler." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Map the UnregisterReply handler.
	if(AM_SetHandler(newEndpoint, S_UNREGISTER_REPLY_HANDLER, (Function)cS_Factory::UnregisterReplyHandler) != AM_OK)
	{
		cout << "Error: Unable to set the unregister reply handler." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Map the Update request handler
	if(AM_SetHandler(newEndpoint, S_UPDATE_HANDLER, (Function)cS_Factory::UpdateHandler) != AM_OK)
	{
		cout << "Error: Unable to set the update reply handler." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Map the return container reply handler
	if(AM_SetHandler(newEndpoint, S_RETURN_CONTAINER_REPLY_HANDLER, (Function)cS_Factory::ReturnContainerReplyHandler) != AM_OK)
	{
		cout << "Error: Unable to set the return container reply handler." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Map the return to sender handler.
	if(AM_SetHandler(newEndpoint, 0, (Function)cS_Factory::ReturnToSenderHandler) != AM_OK)
	{
		cout << "Error: Unable to set the return to sender handler." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Create the reader.
	try
	{
		newReader = new cS_Reader(originalCont, newEndpoint, callbackFunction);
		if(!newReader)
		{
			cout << "Unable to create the new shared reader object in CreateReader." << endl;
			AM_FreeEndpoint(newEndpoint);
			return NULL;
		}
	}
	catch(cShareException &e)
	{
		cerr << e.ToString() << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Insert into reader table.
	if(mReaderTable->Insert((void *)newEndpoint, (void *)newReader) == NULL)
	{
		cout << "Error: Unable to enter new reader into reader table." << endl;
		AM_FreeEndpoint(newEndpoint);
		delete newReader;
		return NULL;
	}

	mCS.Enter();

		mRegisterMutex.Lock();

		mRegistered = false;

		// Attempt to register.
		if(AM_Request4(newEndpoint, 1, S_REGISTER_HANDLER, 0, 0, 0, 0) != AM_OK)
		{
			cout << "Error: Unable to send registration request." << endl;
			AM_FreeEndpoint(newEndpoint);
			return NULL;
		}
		mRegisterMutex.Lock(S_REGISTER_TIMEOUT); // Wait for reply (for a timeout!)
		Sleep(1000);
		mRegisterMutex.Unlock();	// Unlock, just in case.

		// See if reply was successful.
		result = mRegistered;

	mCS.Leave();

	// Return result!
	if(result)
	{
		// send entire container to new reader
		newReader->_UpdateCont();
		return newReader;
	}
	else
	{
		mReaderTable->Delete((void *)newEndpoint);
		AM_FreeEndpoint(newEndpoint);
		delete newReader;
		return NULL;
	}
}

/*
 * cS_Factory::CreateWriter()
 * 
 * Purpose:	Attempt to create a new writer object. 
 * IN:		originalCont	-> pointer to the original container.
 *			regName			-> name to register w/ nameservice under.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The mWriterTable[] has a new writer in it.
 * Return:	A new cS_Writer object or NULL if fail.
 */
cS_Writer* cS_Factory::CreateWriter(cStdCont *originalCont, char *regName)
{
	en_t		newEndpointName;
	ep_t		newEndpoint;
	cS_Writer  *newWriter;

	if(!_IncreaseRefCount())
	{
		return NULL;
	}
	if(AM_AllocateEndpoint(cS_Factory::mBundle, &newEndpoint, &newEndpointName) != AM_OK)
	{
		cout << "Unable to allocate endpoint in CreateWriter." << endl;
		return NULL;
	}
	AM_SetTag(newEndpoint, S_TAG);
	if(AM_SetExpectedResources(newEndpoint, 10, 10) != AM_OK)
	{
		cout << "Error: Unable to allocate enough endpoint resources in CreateWriter." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Create the writer.
	try
	{
		newWriter = new cS_Writer(originalCont, newEndpoint);
		if(!newWriter)
		{
			cout << "Unable to create the new shared writer object in CreateWriter." << endl;
			AM_FreeEndpoint(newEndpoint);
			return NULL;
		}
	}
	catch(cShareException &e)
	{
		cerr << e.ToString() << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Add this writer to the 'writer table'
	mWriterTable->Insert((void *)newEndpoint, (void *)newWriter);

	// Register the handler for registration.
	if(AM_SetHandler(newEndpoint, S_REGISTER_HANDLER, (Function)cS_Factory::RegisterHandler) != AM_OK)
	{
		cout << "Unable to register the Register Handler in new writer." << endl;
		AM_FreeEndpoint(newEndpoint);
		delete newWriter;
		return NULL;
	}

	// Register the handler for unregistration.
	if(AM_SetHandler(newEndpoint, S_UNREGISTER_HANDLER, (Function)cS_Factory::UnregisterHandler) != AM_OK)
	{
		cout << "Unable to register the Unregister Handler in new writer." << endl;
		AM_FreeEndpoint(newEndpoint);
		delete newWriter;
		return NULL;
	}

	// Map the Update reply handler
	if(AM_SetHandler(newEndpoint, S_UPDATE_REPLY_HANDLER, (Function)cS_Factory::UpdateReplyHandler) != AM_OK)
	{
		cout << "Error: Unable to set the update reply handler." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Map the Return Container Request handler
	if(AM_SetHandler(newEndpoint, S_RETURN_CONTAINER_HANDLER, (Function)cS_Factory::ReturnContainerHandler) != AM_OK)
	{
		cout << "Error: Unable to set the container request handler." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Map the return to sender handler.
	if(AM_SetHandler(newEndpoint, 0, (Function)cS_Factory::ReturnToSenderHandler) != AM_OK)
	{
		cout << "Error: Unable to set the return to sender handler." << endl;
		AM_FreeEndpoint(newEndpoint);
		return NULL;
	}

	// Register w/ name service.
	if(AMNS_Register(regName, newEndpointName) != AM_OK)
	{
		cout << "Unable to register endpoint name in CreateWriter." << endl;
		AM_FreeEndpoint(newEndpoint);
		delete newWriter;
		return NULL;
	}

	// Success
	if(!newWriter->_StartWriterThread())
	{
		cout << "Unable to start writer's thread." << endl;
		AM_FreeEndpoint(newEndpoint);
		delete newWriter;
		return NULL;		
	}
	return newWriter;
}

/*
 * cS_Factory::DeleteReader()
 * 
 * Purpose:	Frees a reader object and its resources (but not the original container)
 * IN:		reader	-> The reader to delete.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The mReader table has one less reader in it.
 * Return:	true (always, heh!)
 */
bool cS_Factory::DeleteReader(cS_Reader *reader)
{
	bool	result;

	// Sanity check
	if(!reader)
	{
		return false;
	}

	// Unregister from the writer it is attached to.
	mCS.Enter();

		mUnregisterMutex.Lock();

		mUnregistered = false;

		// Attempt to unregister.
		if(AM_Request4(reader->mEndpoint, 1, S_UNREGISTER_HANDLER, reader->mSlot, 0, 0, 0) != AM_OK)
		{
			cout << "Error: Unable to send unregistration request." << endl;
		}
		mUnregisterMutex.Lock(S_REGISTER_TIMEOUT);	// Wait for reply (for a timeout!)
		mUnregisterMutex.Unlock();					// Unlock, just in case.

		// See if reply was successful.
		result = mUnregistered;

	mCS.Leave();
	if(!result)
	{
		cerr << "Error: Unable to unregister reader!" << endl;
	}
	else
	{
		cerr << "UNREGISTER SUCCESS!" << endl;
	}
	mReaderTable->Delete((void *)reader->mEndpoint);
	AM_FreeEndpoint(reader->mEndpoint);
	delete reader;	// TEMP!

	// Decrease refernce count.
	if(!cS_Factory::_DecreaseRefCount())
	{
		cout << "Unable to reduce ref count in DeleteReader." << endl;
	}
	return true;
}

/*
 * cS_Factory::DeleteWriter()
 * 
 * Purpose:	Frees a writer object and its resources (but not the original container.
 * IN:		writer		-> The writer to delete.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The mWriterTable[] has the writer removed.
 * Return:	true (always, heh!)
 */
bool cS_Factory::DeleteWriter(cS_Writer *writer)
{
	if(!writer)
	{
		return false;
	}
	writer->_SendUpdate(SHARE_CONT_WRITER_GONE);
	writer->_StopWriterThread();
	mWriterTable->Delete((void *)writer->mEndpoint);
	AM_FreeEndpoint(writer->mEndpoint);
	delete writer;

	if(!cS_Factory::_DecreaseRefCount())
	{
		cout << "Unable to reduce ref count in DeleteWriter." << endl;
	}
	return true;
}

/*
 * cS_Factory::_IncreaseRefCount()
 * 
 * Purpose:	Increases the reference count of readers/writers.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	S resources are set up if ref was 0.
 * Return:	true || false
 */
bool cS_Factory::_IncreaseRefCount()
{
	mCS.Enter();

	mRefCount++;

	if(mRefCount == 1)
	{

		// Create data structures.
		mWriterTable = new cHashTable(S_MAX_NUM_WRITERS);
		mReaderTable = new cHashTable(S_MAX_NUM_READERS);
		if(!mReaderTable || !mWriterTable)
		{
			cout << "Error: Unable to alloc hash tables!" << endl;
			return false;
		}

		// Start up AM!
		if(AM_Init() != AM_OK)
		{
			mCS.Leave();
			return false;
		}
		if(AM_AllocateBundle(AM_SEQ, &cS_Factory::mBundle) != AM_OK)
		{
			mCS.Leave();
			return false;
		}
		AM_SetErrorStream(cerr);
	}

	mCS.Leave();
	return true;
}

/*
 * cS_Factory::_DecreaseRefCount()
 * 
 * Purpose:	Decreases the reference count of readers/writers.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	S resources are released if ref goes to zero.
 * Return:	true || false
 */
bool cS_Factory::_DecreaseRefCount()
{
	mCS.Enter();

	mRefCount--;

	if(mRefCount == 0)
	{
		// Shut down AM!
		if(AM_Terminate() != AM_OK)
		{
			mCS.Leave();
			return false;
		}

		// Delete hash tables.
		if(mWriterTable) { delete mWriterTable; }
		if(mReaderTable) { delete mReaderTable; }
	}

	mCS.Leave();
	return true;
}


/*
 * cS_Factory::SetNameServiceNode()
 * 
 * Purpose:	One time deal to set the location of the name service.
 * IN:		node	-> The node on which the name service is to run.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	true || false
 */
bool cS_Factory::SetNameServiceNode(char *node)
{
	return (AMNS_Init(node) == AM_OK);
}


