/*
 * cS_Writer.cpp
 * 
 * Function definitions for shared, writer container.
 *
 */

#include "cS_Reader.h"
#include "cS_Factory.h"

/*
 * cS_Reader::cS_Reader()
 * 
 * Purpose:	Creates a new shared reader object.
 * IN:		originalCont	-> The container that this objects works with.
 *			endpoint		-> This shared container's endpoint.
 *			callback		-> A callback function that can be used when the writer dies (optional).
 * OUT:		-
 * Cond:	-
 * PostCnd:	The reader is created w/ a NULL queue.
 * Return:	AM_OK or AM_ERR_*
 */
cS_Reader::cS_Reader(cStdCont *originalCont, ep_t endpoint, Callback callback)
{
	UpdateCommand	*update;

	mEndpoint		  = endpoint;
	mOrigCont		  = originalCont;
	mFuncQ			  = NULL;
	mFreeQ 			  = NULL;
	mCallback		  = callback;
	mNewCommand		  = NULL;
	mCurrentCommand	  = NULL;
	mWaitForContainer = true;  // First thing wait for is container.
	mWriterDead		  = false;

	// Create function and free queues.
	try
	{
		mFuncQ = new cQueue(MAX_FIFO_SIZE);
		if(!mFuncQ)
		{
			throw cShareException("Error: Unable to create function queue.", __LINE__, __FILE__);
		}
		mFreeQ = new cQueue(MAX_FIFO_SIZE);
		if(!mFreeQ)
		{
			delete mFuncQ;
			mFuncQ = NULL;
			throw cShareException("Error: Unable to create free queue.", __LINE__, __FILE__);
		}
	}
	catch(cQueueException &e)
	{
		cout << e.ToString() << endl;
		if(mFuncQ) { delete mFuncQ; mFuncQ = NULL; }
		if(mFreeQ) { delete mFreeQ; mFreeQ = NULL; }
		throw cShareException("Error: Exception caught in creating Func/Free queues.", __LINE__, __FILE__);
	}

	// Create free updatecommands and populate in free queue.
	for(int i = 0; i < MAX_FIFO_SIZE; i++)
	{
		update = new UpdateCommand;
		if(!update)
		{
			throw cShareException("Error: Unable to create enough update commands.", __LINE__, __FILE__);			
		}
		mFreeQ->Push(update);
	}

	// Create the demarshall stream.
	mDeMarshallStream = new DeMarshallStream(S_MTS);
	if(!mDeMarshallStream)
	{
		throw cShareException("Error: Unable to create demarshallstream.", __LINE__, __FILE__);
	}
}


/*
 * cS_Reader::Insert()
 * 
 * Purpose:	Std container insert operation.
 * IN:		key				-> Some key object.
 *			obj				-> Object to be inserted.
 * OUT:		-
 * Cond:	-
 * PostCnd:	Insert has been called on internal container.
 * Return:	true if success, else false.
 */
bool cS_Reader::Insert(Object* key, Object* obj)
{
	// Perform insert on local object.
	bool check = mOrigCont->Insert(key, obj);
	if(DEBUG && !check) { cout << "Insert failed" << endl; }
	
	// Return true iff writer and readers were successfully updated.
	if (check) { return true;}
	else {return false;}
}

/*
 * cS_Reader::Insert()
 * 
 * Purpose:	Std container delete operation.
 * IN:		key				-> Some key object to base deletion on.
 * OUT:		-
 * Cond:	-
 * PostCnd:	Delete has been called on internal container.
 * Return:	true if success, else false.
 */
bool cS_Reader::Delete(Object* key)
{
	// Perform delete on the local machine.
	bool check = mOrigCont->Delete(key);
	if (DEBUG && !check) { cout << "Delete failed." << endl; }

	// Return true iff writer and readers were successfully updated.
	if (check) { return true;}
	else {return false;}
}

/*
 * cS_Reader::Clear()
 * 
 * Purpose:	Std container clear operation.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	Clear has been called on internal container.
 * Return:	true if success, else false.
 */
bool cS_Reader::Clear()
{
	// Perform clear on the local machine.
	bool check = mOrigCont->Clear(); 
	if (DEBUG && !check) { cout << "Clear failed" << endl; }

	// Return true iff writer and readers were successfully updated.
	if (check) { return true;}
	else {return false;}
}


/*
 * cS_Reader::MakeObjs()
 * 
 * Purpose:	Hack to get internal container to create arguments for us.
 * IN:		func	-> The function that is being referred to (to figure out params)
 * OUT:		arg1	-> A new'd object for argument 1
 *			arg2	-> A new'd object for argument 2
 * Cond:	-
 * PostCnd:	The objects are allocated!
 * Return:	true if success, else false.
 */
bool cS_Reader::MakeObjs(int func, Object** arg1, Object** arg2)
{
	bool check = false;
	*arg1 = NULL;
	*arg2 = NULL;

	if(func == SHARE_CONT_NEWCONT)
	{
		*arg1 = mOrigCont;
		check = true;
	}
	else
	{
		check = mOrigCont->MakeObjs(func, arg1, arg2);
	}
	return check;
}

/*
 * cS_Reader::Marshall()
 * 
 * Purpose:	Turn the original container into a byte array.
 * IN:		buf			-> The buffer in which to marshall stuff.
 *			bufLength	-> The length of the buffer given.
 * OUT:		bufLength	-> The amount of the given buffer used.
 * Cond:	-
 * PostCnd:	The objects are allocated!
 * Return:	true if success, else false.
 */
OBJ_RETURN cS_Reader::Marshall(MarshallStream* stream)
{
	OBJ_RETURN retCode = mOrigCont->Marshall(stream);
	return retCode;
}

/*
 * cS_Reader::DeMarshall()
 * 
 * Purpose:	Not used.
 * IN:		buf			-> The buffer in which to marshall stuff.
 *			bufLength	-> The length of the buffer given.
 * OUT:		bufLength	-> size of demarshalled buf segment
 * Return:	true if success, else false.
 */
OBJ_RETURN cS_Reader::DeMarshall(DeMarshallStream* stream)
{
	// we may only need this function to 
	// realize the virtual function for stdCont.
	// DeMarshalling done in _Update 
	return OBJ_RETURN_FAIL; 
}

/*
 * cS_Reader::GetCont()
 * 
 * Purpose:	Return to the reader a pointer to his container.
 * IN:		-
 * OUT:		cont	-> The address of a pointer to container.
 * Return:	true if success, else fail.
 */
bool cS_Reader::GetCont(cStdCont** cont)
{
	bool retVal;
	if(mWriterDead)
	{
		return false;
	}
	if((GetTickCount() - mTimeStamp) > (S_PING_TIMEOUT + S_READER_TIMEOUT))
	{
		cerr << "<<TIMEOUT EXCEEDED>>: Writer dead?" << endl;
		mWriterDead = true;
		if(mCallback != NULL)
		{
			mCallback();
		}
		retVal = false;
	}
	else
	{
		retVal = true;
	}
	_ProcessCommands();
	*cont = mOrigCont;
	return retVal;
}

/*
 * cS_Reader::_ProcessCommands
 * 
 * Purpose:	Clears the queue by processing all the commands in it.
 * IN:		-
 * Return:	true if success, else false.
 * Note:	If return is false, then request new container to prevent inconsistency.
 */
bool cS_Reader::_ProcessCommands()
{
	UpdateCommand *command;
	Object		  *obj1;
	Object		  *obj2;
	bool		  check = false;
	OBJ_RETURN	  retCode=OBJ_RETURN_FAIL;

	mCS.Enter();

	// Clear out the queue at this point in time.
	
	while(mFuncQ->Peek((void **)&command))
	{
		// Pop off queue.
		mFuncQ->Pop((void **)&command);

		// Make objects for function arguements
		obj1 = NULL;
		obj2 = NULL;
		check = this->MakeObjs(command->opcode, &obj1, &obj2);
		if (!check) { mCS.Leave(); return false; }

		mDeMarshallStream->Start(command);
		if(obj1)
		{
			retCode = obj1->DeMarshall(mDeMarshallStream);
			if (retCode == OBJ_RETURN_FAIL) {return false;}
		}
		if(obj2)
		{
			retCode = obj2->DeMarshall(mDeMarshallStream);
			if (retCode == OBJ_RETURN_FAIL) {return false;}
		}
		mDeMarshallStream->Stop();

		// call appropriate function
		switch (command->opcode){

		  case SHARE_CONT_INSERT: 
			check = this->Insert(obj1, obj2);
			break;
		  case SHARE_CONT_DELETE:
			check = this->Delete(obj1);
			break;
		  case SHARE_CONT_CLEAR:
			check = this->Clear();
			break;
		  case SHARE_CONT_NEWCONT:
			// once we've demarshalled the origCont, we're done
			break;
		  default:
			cout << "Couldn't recognize opcode " << command->opcode << " in Reader Update" << endl;
			break;
		} /* end switch */

		// Push all of the update's buffers back on the free q.
		_FreeCommand(command);
	} /* end while */

	mCS.Leave();
	return true;
}


/*
 * cS_Reader::_Update()
 * 
 * Purpose:	Used to process updates from writer.  
 * IN:		buf		-> buffer containing function args.
 *			nbytes	-> size of buf
 *			opcode	-> function writer performed
 * Return:	retCode = S_RETURN_FAIL or S_RETURN_SUCCESS
 */
int cS_Reader::_Update(void *buf, int nbytes, int opcode)
{
	int retCode = S_RETURN_FAIL;
	bool check = false;
	UpdateCommand *command;

	// Set keep-alive
	_UpdateTimeStamp();

	// Special Case: Writer has left!
	if(opcode == SHARE_CONT_WRITER_GONE)
	{
		cerr << "THE WRITER HAS LEFT!" << endl;
		mWriterDead = true;
		if(mCallback != NULL)
		{
			mCallback();
		}
		return S_RETURN_SUCCESS;
	}
	else if(opcode == SHARE_CONT_PING)
	{
		cerr << "<<PING>>" << endl;
		return S_RETURN_SUCCESS;
	}
	else if(opcode == SHARE_CONT_NEWCONT)
	{
		mWaitForContainer = false;

		// Clear current command.
		if(mCurrentCommand)
		{
			_FreeCommand(mNewCommand);
			mNewCommand		= NULL;
			mCurrentCommand = NULL;
		}

		// Empty queue.
		while(mFuncQ->Pop((void **)&command))
		{
			_FreeCommand(command);
		}
	}

	if(!mWaitForContainer)
	{
		if(mCurrentCommand)
		{
			// We are in the middle of processing a command.
			if(!mFreeQ->Pop((void **)&command))
			{
				// No more free buffers left: Send request for full container.
				_UpdateCont();
			}
			else
			{
				mCurrentCommand->next = command;
				mCurrentCommand = command;
				memcpy(command->buf, buf, nbytes);
				command->opcode = opcode;
				command->next   = NULL;

				if(nbytes < S_MTS)
				{
					if(!mFuncQ->Push(mNewCommand))
					{
						_UpdateCont();
						_FreeCommand(mNewCommand);
					}
					mNewCommand = NULL;
					mCurrentCommand = NULL;
				}
			}
		}
		else
		{
			// New command.
			// We are in the middle of processing a command.
			if(!mFreeQ->Pop((void **)&command))
			{
				// No more free buffers left!
				_UpdateCont();
			}
			else
			{
				mNewCommand = command;
				mCurrentCommand = command;
				command->next = NULL;

				mNewCommand->opcode = opcode;
				memcpy(command->buf, buf, nbytes);

				// If the buffer is not full, then this is a full command.
				if(nbytes < S_MTS)
				{
					if(!mFuncQ->Push(mNewCommand))
					{
						_UpdateCont();
						_FreeCommand(mNewCommand);
					}
					mNewCommand = NULL;
					mCurrentCommand = NULL;
				}
			}
		}
	}

	// return success
	retCode = S_RETURN_SUCCESS;
	return retCode;
}

// destructor
cS_Reader::~cS_Reader()
{
	UpdateCommand *command;
	if(mFuncQ)
	{ 
		while(mFuncQ->Pop((void **)&command))
		{
			delete command;
		}
		delete mFuncQ; 
	}
	if(mFreeQ) 
	{ 
		while(mFreeQ->Pop((void **)&command))
		{
			delete command;
		}
		delete mFreeQ; 
	}
}

/*
 * cS_Reader::_FreeCommand()
 * 
 * Purpose:	Push buffers back onto free queue.  
 * IN:		
 * Return:	success or failure
 */
bool cS_Reader::_FreeCommand(UpdateCommand *command)
{
	if(command->next != NULL)
	{
		_FreeCommand(command->next);
	}
	mFreeQ->Push(command);
	return true;
}

/*
 * cS_Reader::_UpdateCont()
 * 
 * Purpose:	Request entire container.  
 * IN:		
 * Return:	success or failure
 */
OBJ_RETURN cS_Reader::_UpdateCont()
{
	if(AM_Request4(mEndpoint, 1, S_RETURN_CONTAINER_HANDLER, mSlot, 0, 0, 0) != AM_OK)
	{
		cout << "Error: Unable to request entire container." << endl;
		return OBJ_RETURN_FAIL;
	}
	mWaitForContainer = true;

	return OBJ_RETURN_SUCCESS;
}
