/*
 * cS_Writer.cpp
 * 
 * Function definitions for shared, writer container.
 *
 */

#include <am.h>
#include "cS_Writer.h"

/*
 * cS_Writer::cS_Writer()
 * 
 * Purpose:	Creates a new shared writer.
 * IN:		orig_cont	-> The internal container that is to be shared.
 *			endpoint	-> This writer's endpoint.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The shared writer is set up.
 * Return:	-
 */
cS_Writer::cS_Writer(cStdCont* orig_cont, ep_t endpoint)
{
	mMarshallStream	= NULL;
	mMutex			= NULL;
	mWriterThreadQ  = NULL;
	mRun			= false;

	// Set the endpoint.
	mEndpoint = endpoint;

	// Keep a pointer to the original object.
	mOrigCont = orig_cont;

	// Set the insert location for readers array.
	mReadersIndex = 0;

	// Create the marshall stream
	mMarshallStream = new MarshallStream(S_MTS, this);
	if(!mMarshallStream)
	{
		throw cShareException("Error: Unable to create marshall steam.", __LINE__, __FILE__);
	}

	// Create the writer thread's queue.
	try
	{
		mWriterThreadQ = new cQueue(S_MAX_NUM_READERS);
		if(!mWriterThreadQ)
		{
			throw cShareException("Error: Unable to create thread queue.", __LINE__, __FILE__);
		}
	}
	catch(cQueueException &e)
	{
		cerr << e.ToString() << endl;
		throw cShareException("Error: Unable to create thread queue.", __LINE__, __FILE__);
	}
	mMutex = new cMutex(FALSE);
	if(!mMutex)
	{
		throw cShareException("Error: Unable to create thread mutex.", __LINE__, __FILE__);
	}
}

/*
 * cS_Writer::Insert()
 * 
 * Purpose:	Standard container insert.  Mirrored to all readers.
 * IN:		key			-> Some sort of key object.
 *			obj			-> Object to be inserted.
 * OUT:		-
 * Cond:	-
 * PostCnd:	Insert has been called on internal container.
 * Return:	true | false
 */
bool cS_Writer::Insert(Object* key, Object* obj)
{
	// Perform insert on local object.
	bool checkW = mOrigCont->Insert(key, obj);
	if(DEBUG && !checkW) {cout << "Insert failed" << endl;}

	// Send update to reader(s).
	bool checkR = _SendUpdate(SHARE_CONT_INSERT, key, obj);
	if (DEBUG && !checkR){cout << "SendMsg for Insert failed." << endl;}
	
	// Return true iff writer and readers were successfully updated.
	if (checkW && checkR) {	return true;}
	else {return false;}
}

/*
 * cS_Writer::Delete()
 * 
 * Purpose:	Standard container delete.  Mirrored to all readers.
 * IN:		key			-> Some sort of key object.
 * OUT:		-
 * Cond:	-
 * PostCnd:	Delete has been called on internal container.
 * Return:	true | false
 */
bool cS_Writer::Delete(Object* key)
{
	// Perform delete on the local machine.
	bool checkW = mOrigCont->Delete(key);

	if (DEBUG && !checkW){cout << "Delete failed." << endl;}

	// Send update to readers.
	bool checkR = _SendUpdate(SHARE_CONT_DELETE, key);

	if (DEBUG && !checkR){cout << "SendMsg for Delete failed." << endl;}

	// Return true iff writer and readers were successfully updated.
	if (checkW && checkR) {	return true;}
	else {return false;}
}

/*
 * cS_Writer::Clear()
 * 
 * Purpose:	Standard container clear.  Mirrored to all readers.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	Clear has been called on internal container.
 * Return:	true | false
 */
bool cS_Writer::Clear()
{
	// Perform clear on the local machine.
	bool checkW = mOrigCont->Clear();

	if (DEBUG && !checkW){cout << "Clear failed" << endl;}

	// Send update to readers.
	bool checkR = _SendUpdate(SHARE_CONT_CLEAR);
	
	if (DEBUG && !checkR){cout << "SendMsg for Clear failed." << endl;}

	// Return true iff writer and readers were successfully updated.
	if (checkW && checkR) {	return true;}
	else {return false;}
}

/*
 * cS_Writer::_SendContainer()
 * 
 * Purpose:	Sends the entire container to client | ALL
 * IN:		client		-> Destination to send to.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	true | false
 */
bool cS_Writer::_SendContainer(int client)
{
	return _SendUpdate(SHARE_CONT_NEWCONT, mOrigCont, NULL, client);
}

/*
 * cS_Writer::MakeObjs()
 * 
 * Purpose:	Hack to create arguments for function calls.
 * IN:		func		-> The function to create args for.
 * OUT:		arg1		-> New'd arg1 object.
 *			arg2		-> New'd arg2 object.
 * Cond:	-
 * PostCnd:	-
 * Return:	true | false
 */
bool cS_Writer::MakeObjs(int func, Object** arg1, Object** arg2)
{
	bool check = mOrigCont->MakeObjs(func, arg1, arg2);
	return check;
}

/*
 * cS_Writer::Marshall()
 * 
 * Purpose:	Marshalls the internal container.
 * IN:		buf			-> A buffer to marshall into.
 *			bufLength	-> Total size of buffer.
 * OUT:		bufLength	-> The amount of buffer used.
 * Cond:	-
 * PostCnd:	-
 * Return:	true | false
 */
OBJ_RETURN cS_Writer::Marshall(MarshallStream* stream)
{
	OBJ_RETURN retCode = mOrigCont->Marshall(stream);
	return retCode;
}

/*
 * cS_Writer::DeMarshall()
 * 
 * Purpose:	Not implemented.
 * IN:		buf			-> A buffer to demarshall from.
 *			bufLength	-> Total size of buffer.
 * OUT:		bufLength	-> size of demarshalled buf segment
 * Cond:	-
 * PostCnd:	-
 * Return:	true | false
 */
OBJ_RETURN cS_Writer::DeMarshall(DeMarshallStream* stream)
{
	// not currently being used
	return OBJ_RETURN_FAIL;
}

/*
 * cS_Writer::_SendUpdate()
 * 
 * Purpose:	Sends the update to all registered clients.
 * IN:		func		-> The function opcode.
 *			arg1		-> First argument to marshal
 *			arg2		-> Second argument to marshal
 *			client		-> Who to send to (1 client or _ALL)
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	true | false
 */
bool cS_Writer::_SendUpdate(int func, Object* arg1, Object* arg2, int client)
{
	mCSUpdate.Enter();
	mMarshallStream->StartArgs(func, client);
	if(arg1 != NULL)
	{
		arg1->Marshall(mMarshallStream);
	}
	if(arg2 != NULL)
	{
		arg2->Marshall(mMarshallStream);
	}
	mMarshallStream->StopArgs();
	
	mCSUpdate.Leave();
	return true;
}


/*
 * cS_Writer::_SendMsgAll()
 * 
 * Purpose:	Sends the given message to all registered clients.
 * IN:		opcode		-> The function opcode.
 *			buffer		-> The buffer to send.
 *			nbytes		-> The number of bytes to send from buffer.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	true | false
 */
bool cS_Writer::_SendMsgAll(int opcode, char *buffer, int nbytes)
{
	int rc;

	// Use AM to send buf
	mCSReaders.Enter();	 // Enter reader array
	for(int k = 0; k < mReadersIndex; k++)
	{	
		rc = AM_RequestI4(mEndpoint, mReaders[k].index, S_UPDATE_HANDLER, buffer, nbytes, opcode, 0, 0, 0);
		if(rc != AM_OK)
		{
			if(rc == AM_ERR_RESOURCE)
			{
				// Client disconnected.
				cerr << "Client " << k << " disconnected.  Removing." << endl;
				AM_Unmap(mEndpoint, mReaders[k].index);
				_UnregisterReader(mReaders[k].index);
			}
			else if(rc == AM_ERR_NOT_SENT)
			{
				cerr << "Error: Unable to send active message to reader: " << mReaders[k].index << endl;
				cerr << "Send window full?" << endl;
			}
		}
		else
		{
			cout << "Sent message to reader " << mReaders[k].index << endl;
		}
	}
	mCSReaders.Leave();
	_UpdateTimeStamp();
	return true;
}

/*
 * cS_Writer::_SendMsg()
 * 
 * Purpose:	Sends the given message to the given client.
 * IN:		opcode		-> The function opcode.
 *			buffer		-> The buffer to send.
 *			nbytes		-> The amount to send from the buffer.
 *			client		-> The single client to send to.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	true | false
 */
bool cS_Writer::_SendMsg(int opcode, char *buffer, int nbytes, int client)
{
	int rc;

	// Use AM to send buf
	mCSReaders.Enter();	 // Enter reader array

	if(mReaders[client].valid == false)
	{
		mCSReaders.Leave();
		return false;
	}

	rc = AM_RequestI4(mEndpoint, client, S_UPDATE_HANDLER, buffer, nbytes, opcode, 0, 0, 0);
	if(rc != AM_OK)
	{
		if(rc == AM_ERR_RESOURCE)
		{
			// Client disconnected.
			cerr << "Client " << client << " disconnected.  Removing." << endl;
			AM_Unmap(mEndpoint, client);
			_UnregisterReader(client);
		}
		else if(rc == AM_ERR_NOT_SENT)
		{
			cerr << "Error: Unable to send active message to reader: " << client << endl;
			cerr << "Send window full?" << endl;
		}
	}
	else
	{
		cout << "Sent message to reader " << client << endl;
	}
	mCSReaders.Leave();
	return true;
}


/*
 * cS_Writer::_RegisterReader()
 * 
 * Purpose:	Registers a reader w/ this writer.
 * IN:		newReader	-> Address of new reader to connect to.
 * OUT:		slot		-> The index into the map table for this reader.
 * Cond:	-
 * PostCnd:	The reader is registered w/ writer and will receive updates.
 * Return:	true | false
 */
bool cS_Writer::_RegisterReader(en_t newReader, int *slot)
{
	int		translationSlot;
	int		rc;
	bool	retVal;

	mCSReaders.Enter();

	if(mReadersIndex == MAX_NUM_READERS)
	{
		retVal = false;	// No room in our array.
	}
	else
	{
		rc = AM_MapAny(mEndpoint, &translationSlot, newReader, S_TAG);
		if(rc != AM_OK)
		{
			retVal = false;	// Unable to map endpoint.
		}

		mReaders[mReadersIndex].index = translationSlot;
		mReaders[mReadersIndex].valid = true;
		*slot = translationSlot;
		mReadersIndex++;
		retVal = true;
	}
	
	mCSReaders.Leave();

	return retVal;
}

/*
 * cS_Writer::_UnregisterReader()
 * 
 * Purpose:	Removes the reader from the eps translation table.
 * IN:		reader		-> The value in the reader array to search for.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The reader array has the given reader unregistered.
 * Return:	AM_OK or AM_ERR_*
 */
bool cS_Writer::_UnregisterReader(int reader)
{
	mCSReaders.Enter();

	if(mReadersIndex == 0)
	{
		mCSReaders.Leave();
		return false;
	}

	// Search for the given reader.
	for(int i = 0; i < mReadersIndex; i++)
	{
		if(mReaders[i].index == reader)
		{
			mReaders[i].valid = false;
			
			// Try to compact.
			mReadersIndex--;
			if(mReaders[mReadersIndex].valid)
			{
				mReaders[i].index = mReaders[mReadersIndex].index;
				mReaders[i].valid = true;
				mReaders[mReadersIndex].valid = false;
			}
			mCSReaders.Leave();
			return true;
		}
	}

	// Not found.
	mCSReaders.Leave();
	return false;
}

/*
 * cS_Writer::_UpdateReply()
 * 
 * Purpose:	Handles an update reply from a reader.
 * IN:		retCode		-> The reader's return code.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	true | false
 */
bool cS_Writer::_UpdateReply(int retCode)
{
// At this point, reader has reported failure.
	return true;
}

/*
 * cS_Writer::~cS_Writer()
 * 
 * Purpose:	Destroys this writer object.
 */
cS_Writer::~cS_Writer()
{
	if(mMarshallStream)
	{
		delete mMarshallStream;
	}
}

/*
 * cS_Writer::_Compact()
 * 
 * Purpose:	Compacts the reader array for more efficient sending.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	The reader array has no empty gaps in it.
 * Return:	AM_OK or AM_ERR_*
 */
bool cS_Writer::_Compact()
{
	int	i;

	mCSReaders.Enter();

	for(i = 0; i < mReadersIndex; i++)
	{
		if(!mReaders[i].valid)
		{
			// We know there is a hole.
			while(mReadersIndex > i)
			{
				if(mReaders[mReadersIndex].valid)
				{
					mReaders[i].index = mReaders[mReadersIndex].index;
					mReaders[i].valid = true;
					mReaders[mReadersIndex].valid = false;
					break;
				}
				else
				{
					mReadersIndex--;
				}
			}
		}
	}

	mCSReaders.Leave();
	return true;
}


/*
 * cS_Writer::WriterThread()
 *
 * Purpose:	Main routine for the writer thread.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	Will execute until mRun is set to false.
 * Return:	Proper thread exit return code.
 */
DWORD cS_Writer::_WriterThreadMain()
{
	DWORD	retCode = 0;
	int		client;

	_UpdateTimeStamp();
	
	while(mRun)
	{
		mMutex->Lock(S_PING_TIMEOUT);
//		Sleep(S_PING_TIMEOUT);

		while(mWriterThreadQ->Pop((void **)&client))
		{
			_SendContainer(client);
		}

		if((GetTickCount() - mTimeStamp) > S_PING_TIMEOUT)
		{
			_SendUpdate(SHARE_CONT_PING);
			_UpdateTimeStamp();
		}
	}
	return retCode;
}


/*
 * cS_Writer::WriterThread()
 *
 * Purpose:	The thread entity that will call _WriterThreadMain.
 * IN:		A pointer to the cS_Writer object.
 * OUT:		-
 * Cond:	A thread has been created w/ entry point WriterThread.
 * PostCnd:	Will wait until return from _WriterThreadMain.
 * Return:	Proper thread exit return code.
 */
DWORD WINAPI cS_Writer::WriterThread(LPVOID lpParameter)
{
	DWORD		 retCode = 0;
	cS_Writer	*writer;
	printf("Inside of thread...whoopee!\n");
	
	//Start up the bundle
	writer = (cS_Writer *)lpParameter;
	retCode = writer->_WriterThreadMain();

	//Shut down the thread.
	printf("WriterThread(): Shutting down thread.\n");
	ExitThread(retCode);

	//To avoid warning
	return retCode;
}


/*
 * cS_Writer::_StartWriterThread()
 *
 * Purpose:	Starts the writer's thread.
 * IN:		-
 * OUT:		-
 * Cond:	The WriterThread is not running..
 * PostCnd:	The WriterThread should be running.
 * Return:	true if success, else false
 */
bool  cS_Writer::_StartWriterThread()
{
	// Make sure it is not already started.
	if (mRun)
	{
		return false;
	}

	// Thread settings vars
	LPSECURITY_ATTRIBUTES	lpThreadAttributes;
	DWORD					dwStackSize;
	LPTHREAD_START_ROUTINE	lpStartAddress;
	LPVOID					lpParameter;
	DWORD					dwCreationFlags;
	LPDWORD					lpThreadId;

	// Allow threads to run
	mRun = true;

	// Set up the main thread settings.
	lpThreadId			= &mWriterThreadId;
	dwCreationFlags		= NULL;
	lpParameter			= this;	//Address of this object.		
	dwStackSize			= 0;	//Same as starting parent.
	lpStartAddress		= WriterThread;
	lpThreadAttributes	= NULL;

	// Start the main thread.
	mWriterThreadHandle = CreateThread(
		lpThreadAttributes,  // pointer to security attributes
		dwStackSize,         // initial thread stack size
		lpStartAddress,      // pointer to thread function
		lpParameter,         // argument for new thread
		dwCreationFlags,     // creation flags
		lpThreadId           // pointer to receive thread ID
	);
	if(mWriterThreadHandle == NULL)
	{
		mRun = false;
		return mRun;
	}

	return mRun;
}


/*
 * cS_Writer::_StopWriterThread()
 *
 * Purpose:	Waits for the writer thread to finish up.
 * IN:		-
 * OUT:		-
 * Cond:	The WriterThread is executing.
 * PostCnd:	The WriterThread will have terminated.
 * Return:	true if success, else hangs forever.
 */
bool cS_Writer::_StopWriterThread()
{
	DWORD	retCode = STILL_ACTIVE;

	mRun = false;

	while(retCode == STILL_ACTIVE)
	{
		Sleep(15);
		GetExitCodeThread(mWriterThreadHandle, &retCode);
	}
	printf("Stop(): Writer Thread done.\n");
	return true;
}