/* 
 * cS_Writer.h
 * Class definition for shared writer functions.
 *
 * Rather than updating a user-defined object, the user calls a function in cS_Writer.
 * When a cS_Writer function is called, the user's object is updated and the update
 * is sent to all registered readers.
 *
 */

#ifndef __C_SWRITER_H
#define __C_SWRITER_H

#include "stdrw.h"				// standard header files
#include "cStdCont.h"			// shared container interface
#include <am.h>					// active messaging
#include "cCriticalSection.h"	// critical section
#include "MarshallStream.h"		// Marshall Stream to reader(s).
#include "cQueue.h"				// Queue for writer thread.
#include "cMutex.h"				// Used to wake up writer thread.

#define	MAX_NUM_READERS 128

typedef struct 
{
	int		index;
	bool	valid;
} ReaderEntry;

// cS_Writer Class
class cS_Writer : public cStdCont
{
	public:
		friend class cS_Factory;
		friend class MarshallStream;
		
		//Constructor(s)
		cS_Writer(cStdCont* orig_cont, ep_t endpoint);

		//Functions to update the shared object.
		bool Insert(Object* key, Object* obj);
		bool Delete(Object* key);
		bool Clear();

		// Used for deMarshilling.  Given a function, MakeObjs creates
		// the necessary objects to call the function on the Reader.
		bool MakeObjs(int func, Object** arg1, Object** arg2);
		
		OBJ_RETURN Marshall(MarshallStream* stream);
		OBJ_RETURN DeMarshall(DeMarshallStream* stream); 

		//Destructor
		~cS_Writer(); 

	private:

		// Send AM to Reader(s)
		bool _SendMsgAll(int opcode, char *buffer, int nbytes);
		bool _SendMsg(int opcode, char *buffer, int nbytes, int client);
		bool _SendUpdate(int func, Object* arg1=NULL, Object* arg2=NULL, int client=_ALL);

		bool _Compact();

		//Functions to maintain data on registered Reader(s)
		bool _RegisterReader(en_t newReader, int *slot);
		bool _UnregisterReader(int reader);
		bool _UpdateReply(int retCode);
		bool _SendContainer(int client);
		void _UpdateTimeStamp() { mTimeStamp = GetTickCount(); }

		// Register self with name service
//		bool Register(char* descr, entPTR endpt);

		//Member Vars
		cStdCont*		mOrigCont; // pointer to writer's original struct.
		ep_t			mEndpoint; // This writer's endpoint.
		ReaderEntry		mReaders[MAX_NUM_READERS];
		int				mReadersIndex;
		MarshallStream *mMarshallStream;
		DWORD			mTimeStamp;

		// Locks
		cCriticalSection	mCSReaders;
		cCriticalSection	mCSUpdate;

		// Thread
		HANDLE		  mWriterThreadHandle;	// thread handle of the writer thread.
		DWORD		  mWriterThreadId;		// id of the writer thread.
		static DWORD WINAPI WriterThread(LPVOID lpParameter);
		bool  _StartWriterThread();
		bool  _StopWriterThread();
		DWORD _WriterThreadMain();
		volatile bool mRun;
		cQueue		 *mWriterThreadQ;
		cMutex		 *mMutex;

	protected:

};

#endif